#include "cpsp/HCore.hh"
#include <ell/assertell.hh>

namespace cpsp
{
	
	HCore::HCore() :
		contacts(0), size(0), numOfPSinglets(-1), numOfHSinglets(-1),
		evenPositions(0)
	{
		coreFrame.reset();	// coreFrame initialisieren
	}

	HCore::HCore(const unsigned int contacts_, const unsigned int size_) :
		contacts(contacts_), size(size_), numOfPSinglets(-1), 
		numOfHSinglets(-1), evenPositions(0)
	{
		coreFrame.reset();	// coreFrame initialisieren
	}
	
	HCore::~HCore()
	{
	}
	
	void 
	HCore::reset(const unsigned int contacts_, const unsigned int size_) {
		contacts = contacts_;
		size = size_;
		points.clear();
		coreFrame.reset();
		actFrameSize = 0;
		numOfPSinglets = -1;
		numOfHSinglets = -1;
		evenPositions = 0;
	}
	
	void 
	HCore::addPoint(const CorePoint& toAdd) {
		points.insert(toAdd);			// einfuegen
		coreFrame.resize(toAdd);		// core frame anpassen wenn noetig
		if (toAdd.isEven())				// pruefen ob gerade
			evenPositions++;
	}	
	
	void 
	HCore::indexHulls(	const ell::LatticeFrame* latFrame, 
						const unsigned int maxHullLevel) {
assertell(latFrame != NULL, "tried to index the hulls without LatticeFrame.");
		
		if ( actFrameSize != latFrame->getFrameSize() ) 
		{	
				// dann neuindizierung noetig
			actFrameSize = latFrame->getFrameSize();
			hull.clear();			// neu indizierung noetig
			numOfPSinglets = -1;	// neu indizierung auch fuer singlets noetig
			numOfHSinglets = -1;
		}
			// abbruch wenn nichts zu tun ist
		if ( hull.size() > maxHullLevel)
			return;
			// noetiger verschiebungs vektor um das coreFrame Zentrum 
			// auf das latticeFrame Zentrum zu verschieben
		ell::IntPoint shift = latFrame->getCenter() - coreFrame.getCenter();
			// kern indizierung noetig ?
		if (hull.size() == 0) {
			hull.push_back(ell::IndexSet());
			for (	CPointSet::const_iterator it = points.begin(); 
					it != points.end(); it++) {
				// punkt verschieben und index einfuegen
				hull[0].insert(latFrame->getIndex((*it) + shift));	
			}
		}
			// indizierte nachbarn fuer schnellere huellen berechnung
		ell::IndexVec neighInd = latFrame->getIndexedNeighborhood();
			// weitere huellen indizieren, wenn noch nicht vorhanden
			// huelle (i+1) enthaelt alle punkte um huelle (i) 
			// INKLUSIVE huelle (i) die nicht im kern liegen
		for (unsigned int lev = hull.size(); lev <= maxHullLevel; lev++) {
			if (lev>1) 	// vorherige huelle einfuegen aber nicht den kern selbst
				hull.push_back(ell::IndexSet(hull[lev-1]));
			else		// neue leere huellenset
				hull.push_back(ell::IndexSet());
				// fuer alle indizes aus der vorherigen huelle (lev-1)
			for (	ell::IndexSet::const_iterator it = hull[lev-1].begin(); 
					it != hull[lev-1].end(); it++) {
					// fuer alle nachbarschaftsvektoren
				for (ell::IndexVec::size_type i = 0; i<neighInd.size(); i++) {
						// wenn nicht im kern
					if (hull[0].find((*it) + neighInd[i]) == hull[0].end()) {
							// fuege kernpos+nachbarVec in huelle ein
						hull[lev].insert((*it) + neighInd[i]);	
					}
				}
			}
		}
	} // indexHulls(..)	
	
	const ell::IndexSet& 
	HCore::getIndexedHull(	const ell::LatticeFrame* latFrame, 
							const unsigned int hullLevel) {
		indexHulls(latFrame, hullLevel);
		return hull[hullLevel];
	} // getIndexedHull(..)
	
	int
	HCore::getNumOfHSinglets(const ell::LatticeFrame* latFrame) {
		if (!isValid())
			numOfHSinglets = -1;
		else {
			if (	actFrameSize != latFrame->getFrameSize() 
					|| numOfHSinglets == -1) 	// dann neuberechnung noetig
			{
				ell::IndexSet	core = getIndexedHull(latFrame, 0),
								hull = getIndexedHull(latFrame, 1);
				ell::IndexVec neighInd = latFrame->getIndexedNeighborhood();
				ell::IndexVec::const_iterator neigh;
				unsigned int n=0;
					// init
				numOfHSinglets = 0;
				hSinglets.clear();
					// gehe alle moeglichen H-singlet positionen des kerns durch
				for (	ell::IndexSet::const_iterator it = core.begin(); 
						it!= core.end();
						it++) {
						// init
					n = 0;
						// pruefe ob min. 2 nachbarn in erster huelle
					for (neigh=neighInd.begin(); neigh != neighInd.end(); neigh++) {
						if ( hull.find((*it)+(*neigh)) != hull.end())
							n++;
						if (n>1) {	// min 2 nachbarn im kern
							numOfHSinglets++;	// habe p-singlet gefunden
							hSinglets.insert(*it);	// merken
							break;				// kann naechste pos. anschaun
						}
					}
				}
			}
		}
		return numOfHSinglets;
	}
	
	int
	HCore::getNumOfPSinglets(const ell::LatticeFrame* latFrame) {
		if (!isValid())
			numOfPSinglets = -1;
		else {
			if (	actFrameSize != latFrame->getFrameSize() 
					|| numOfPSinglets == -1) 	// dann neuberechnung noetig
			{
				ell::IndexSet	core = getIndexedHull(latFrame, 0),
								hull = getIndexedHull(latFrame, 1);
				ell::IndexVec neighInd = latFrame->getIndexedNeighborhood();
				ell::IndexVec::const_iterator neigh;
				unsigned int n=0;
					// init
				numOfPSinglets = 0;
				pSinglets.clear();
					// gehe alle moeglichen P-singlet positionen der huelledurch
				for (	ell::IndexSet::const_iterator it = hull.begin(); 
						it!= hull.end();
						it++) {
						// init
					n = 0;
						// pruefe ob min. 2 nachbarn im kern
					for (neigh=neighInd.begin(); neigh != neighInd.end(); neigh++) {
						if ( core.find((*it)+(*neigh)) != core.end())
							n++;
						if (n>1) {	// min 2 nachbarn im kern
							numOfPSinglets++;	// habe p-singlet gefunden
							pSinglets.insert(*it);	// merken
							break;				// kann naechste pos. anschaun
						}
					}
				}
			}
		}
		return numOfPSinglets;
	}
	
		//! Returns the indexed P-singlet positions of the first hull,
		//! shifted to the center of the lattice frame.
	const ell::IndexSet& 
	HCore::getIndexedPSinglets(const ell::LatticeFrame* latFrame) {
		getNumOfPSinglets(latFrame);
		return pSinglets;
	}

	const ell::IndexSet& 
	HCore::getIndexedHSinglets(const ell::LatticeFrame* latFrame) {
		getNumOfHSinglets(latFrame);
		return hSinglets;
	}
	
	unsigned int 
	HCore::getMinEvenOdd(void) {
		return ( (points.size()-evenPositions) > evenPositions) ?
					evenPositions :
					points.size()-evenPositions;
	}

} // namespace cpsp
