#include "cpsp/HCoreDatabaseFILE.hh"
#include "cpsp/Exception.hh"

#include <algorithm>	// fuer <sort> template
#include <functional>	// fuer <greater> template

#ifdef WIN32			// fuer systemspezifische verzeichnis-inhaltsauflistung
#	include <io.h>
#else
#	include <dirent.h>
#endif

#include <biu/util/Util_String.h>

#include <ell/LatticeDescriptorCUB.hh>
#include <ell/LatticeDescriptorFCC.hh>

namespace cpsp
{

	//////////////////////////////////////////////////////////////////////
	// Konstanten
	//////////////////////////////////////////////////////////////////////
	
	const std::string ROOT_CUB = std::string("Cubic/");
	const std::string ROOT_FCC = std::string("Fcc/");
	
	const std::string PREF_CUB = std::string("cub.");
	const std::string PREF_FCC = std::string("fcc.");
	
	const std::string POSTFIX = std::string(".txt");
	

	HCoreDatabaseFILE::HCoreDatabaseFILE(const std::string& dbRootPath) :	
		coreSize(0),  minHH(0), maxHH(UINT_MAX-2), 
			// check for '/' char at the end of the path
		DBROOTPATH(dbRootPath+(dbRootPath[dbRootPath.size()-1] == '/'?"":"/")),
		actFile(NULL), actFileNamePref(""), actFileHHcontInd(-1)
	{
		actHHcontacts.clear();
	}
	
	HCoreDatabaseFILE::HCoreDatabaseFILE(const HCoreDatabaseFILE& toCopy) :
		coreSize(0),  minHH(toCopy.minHH), maxHH(toCopy.maxHH),
		DBROOTPATH(toCopy.DBROOTPATH), actFile(NULL), actFileNamePref(""),
		actFileHHcontInd(-1)
	{
	}
	
	HCoreDatabaseFILE::~HCoreDatabaseFILE()
	{
		if (actFile != NULL) {
			actFile->close();
			delete actFile;
		}
		actFile = NULL;
	}


	//////////////////////////////////////////////////////////////////////
	// initialisiert datenbank dateizugriff
	//////////////////////////////////////////////////////////////////////
	
	bool HCoreDatabaseFILE::initCoreAccess(const ell::LatticeDescriptor& latDescr, const unsigned int size,
					const unsigned int minHH_, const unsigned int maxHH_)
	{

	// (1) daten initialisieren
		this->coreSize = size;
		this->minHH = minHH_;
		this->maxHH = maxHH_;
	
		if (actFile != NULL) { delete actFile; actFile = NULL; }
		actFileHHcontInd = -1;
		actFileNamePref = "";
	
	// (2) verfuegbare energiedateien pruefen und <actEnergies> setzen
	
		// pfadname zusammenstellen
		std::string actDBdirName = DBROOTPATH, prefix = "";
		// lattice root + size dirName
		if (latDescr == ell::LatticeDescriptorCUB()) {
			actDBdirName += ROOT_CUB; prefix += PREF_CUB;
		} else if (latDescr == ell::LatticeDescriptorFCC()) {
			actDBdirName += ROOT_FCC; prefix += PREF_FCC;
		} else {
			std::cerr <<"\n\tError: No core files for lattice type '"+latDescr.getName()+"' available!\n\n";
			return false;
		}
		actDBdirName += (biu::util::Util_String::int2str(size))+"/";	// verzeichnis fuer kerngroesse
		prefix += (biu::util::Util_String::int2str(size)) + ".";	// prefix der gesuchten kerndateien
	
		actHHcontacts.clear();					// alte daten loeschen
	
		std::string filename;					// eintragsname
	
		// verzeichnis oeffnen und eintraege durchlaufen
	#ifdef WIN32
		long hFile;
		struct _finddata_t c_file;				// verzeichniseintrag
		hFile = _findfirst ((actDBdirName+prefix+"*"+POSTFIX).c_str(), &c_file);	// oeffnen
		if (hFile != -1L) {						// dann verzeichnis vorhanden
	
			do {
				filename = c_file.name;
	#else	// linux
		DIR *actDBdir = opendir(actDBdirName.c_str());	// verzeichnis oeffnen
		if(actDBdir != NULL)					// dann verzeichnis vorhanden
		{
	
			struct dirent *dirEntry;			// verzeichniseintrag
	
			while((dirEntry=readdir(actDBdir)))
			{
				filename = dirEntry->d_name;
	#endif
				// energie der datei bestimmen (hier wirds aus dem dateinamen extrahiert)
				if (filename.substr(0,prefix.size()).compare(prefix) == 0) { // dann ordentliche datei
					filename = filename.erase(0,prefix.size());	// schneide prefix ab
					actHHcontacts.push_back(biu::util::Util_String::str2int(filename.erase(filename.find(POSTFIX)))); // schneide postfix ab und konvertiere
				}
			}
	#ifdef WIN32
			while( _findnext( hFile, &c_file ) == 0 );
	#endif
		} else {
			std::cerr	<<"\n\tError: No core file for size '"<<size <<"' and\n\t       lattice type '" 
						<<latDescr.getName() <<"' available!\n\t==> no directory '"<<actDBdirName<<"'\n\n";
			return false;
			cpsp::Exception::errorCount++;
		}
	
		// verzeichnis schliessen
	#ifdef WIN32
		_findclose( hFile );
	#else
		closedir(actDBdir);
	#endif
	
	// (3) absteigend sortieren
	
		std::sort(actHHcontacts.begin(), actHHcontacts.end()  , std::greater<int>());
	
	// (4) <actFileEnergyInd> setzen und datei oeffnen wenn vorhanden
	
			 // laufe durch bis datei fuer maxenergie gefunden
		unsigned int i=0;
		while ( i<actHHcontacts.size() && actHHcontacts[i]>0 &&((unsigned int)actHHcontacts[i]>maxHH || (unsigned int)actHHcontacts[i]<minHH)) {
//			std::cerr <<" actHH[i] = " <<(int)actHHcontacts[i] <<" size = " <<size <<" minHH = " <<minHH <<" maxHH = " <<maxHH <<"\n";
			i++;
		}
	
		if (i >= actHHcontacts.size()) {// dann KEINE entsprechende datei vorhanden
//			std::cerr <<"no files available\n";
			return false;	
		}
	
		actFileHHcontInd = i;
		actFileNamePref = actDBdirName+prefix;
		if (actFile != NULL) { 
			delete actFile; actFile = NULL;
		}
		actFile = new std::ifstream((actFileNamePref+(biu::util::Util_String::int2str(actHHcontacts[actFileHHcontInd]))+POSTFIX).c_str());
		
		return true;
	} // initCoreAccess()

	bool HCoreDatabaseFILE::getNextCore(HCore& toFill) {
	
		toFill.reset(INT_MAX, INT_MAX);		// ungueltiger dummy kern
		if (actFile == NULL) // nur Kern lesen wenn initialisiert und somit datei schon offen
			return false;
	
		std::string dataline = "";
		int i=0;
		int p[3];

		do {	////////////////////////////////////
				// LESE NAECHSTEN KERN EIN
				////////////////////////////////////
			while (std::getline(*actFile, dataline, '#')) { // datenbloecke durch '#' getrennt
				if (!dataline.empty() && (dataline.substr(0,8).compare("COREDATA")==0)){	// nach schluesselwort 'COREDATA' folgen kerndaten
					try {
						std::istringstream iss(dataline.substr(9)); // reststring in stringstream
						toFill.reset(actHHcontacts[actFileHHcontInd], coreSize); // neuer kern
						for (i=0; i<(int)coreSize; i++) {
							p[2] = INT_MIN; // init zur fehleranalye
							iss >>p[0] >>p[1] >>p[2]; // kernposition auslesen
							if (p[2] == INT_MIN) // dann keine vollstaendigen punktkoordinaten vorhanden
								throw ExcEOF();
							toFill.addPoint(HCore::CorePoint(p[0], p[1], p[2])); // pos zum kern hinzufuegen
						}
					} catch (ExcEOF) // catch block sorgt dafuer dass keine unvollstaendigen daten in <cores> eingefuegt werden
					{
						toFill.reset(INT_MAX,INT_MAX);
						return false;
					}
					return toFill.isValid();
				}
			}
			actFile->close();		// schliessen da alles gelesen
			delete actFile;			// aufraeumen
			actFile = NULL;
	
			// pruefen ob kerne mit geringerer energie noch zu betrachten
			if ((++actFileHHcontInd) < (int)actHHcontacts.size()  && actHHcontacts[actFileHHcontInd]>=(int)minHH) {
				// wenn ja dann diese datei oeffnen
				actFile = new std::ifstream((actFileNamePref+(biu::util::Util_String::int2str(actHHcontacts[actFileHHcontInd]))+POSTFIX).c_str());
			}
		} while (actFile != NULL);
	
		return false;
	}


} // namespace cpsp
